"""
Some useful methods and functions for windowing operations.
"""

import scipy.signal as sig
from numpy import pad, apply_along_axis
from numpy.lib.stride_tricks import as_strided


def apply_sliding_window(x, window_size, hop_size, function, window_name):
    """
    Apply a sliding window with the given parameters to the array `x` and
    aggregate the data within each window using the specified `function`.

    Args
    ----
    x : ndarray
        Array over which the sliding operation must be applied.
    window_size: int
        Size of the window.
    hop_size: int
        Number of samples to be skipped between two successive windowing
        operations.
    window_name : str
        Name of the window function to be used. Options are: {"boxcar",
        "triang", "blackman", "hamming", "hann", "bartlett", "flattop",
        "parzen", "bohman", "blackmanharris", "nuttall", "barthann",
        "no_window", None}.

    Returns
    -------
    y : ndarray
    """
    view = get_sliding_window_view(x, window_size, hop_size)

    if window_name:
        window = get_window_function(window_size, window_name)
        y = apply_along_axis(function, -1, view * window)
    else:
        y = apply_along_axis(function, -1, view)

    return y.T


def get_sliding_window_view(x, window_size, hop_size):
    """
    Generate a view of the input array containing the sliding windows obtained
    for the given parameters.

    This method only creates a view of the sliding windows; it does not apply a
    window function (apodization function) to them.

    Args
    ----
    x : ndarray
        Array over which the sliding operation must be applied.
    window_size: int
        Size of the window.
    hop_size: int
        Number of samples to be skipped between two successive windowing
        operations.

    Returns
    -------
    view : ndarray
    """
    if (x.ndim != 1):
        raise ValueError("`x` must be a unidimensional array")

    pre_padding_size, post_padding_size, num_hops = \
        calculate_sliding_window_parms(window_size, hop_size, x.size)

    # padded version of the array
    x = pad(x, (pre_padding_size, post_padding_size), 'constant',
            constant_values=(0, 0))

    # apply striding tricks to create a windowed view of the array
    view = as_strided(x, shape=(num_hops, window_size),
                      strides=(x.itemsize * hop_size, x.itemsize))

    return view


def get_window_function(window_size, window_name, symmetric=True):
    """
    Get a window function (also known as tapering function or apodization
    function) according to the specified `window_name`.

    This function will return None if the specified `window_name` is also None.

    Args
    ----
    window_name : str
        Name of the window function to be used. Options are: {"boxcar",
        "triang", "blackman", "hamming", "hann", "bartlett", "flattop",
        "parzen", "bohman", "blackmanharris", "nuttall", "barthann",
        "no_window"}.
    """
    # check if the window_name is valid
    possible_windows = {
        "boxcar", "triang", "blackman", "hamming", "hann", "bartlett",
        "flattop", "parzen", "bohman", "blackmanharris", "nuttall",
        "barthann", "no_window"
    }

    if window_name not in possible_windows:
        raise ValueError('invalid window_name: {}'.format(window_name))
    elif window_name is not None:
        return sig.get_window(
            window_name, window_size, fftbins=not symmetric)
    else:
        return None


def calculate_sliding_window_parms(window_size, hop_size, array_size):
    """
    Calculate some parameters that are necessary for applying the sliding
    window over a time series.

    Args
    ----
    window_size: int
        Size of the window.
    hop_size: int
        Number of samples to be skipped between two successive windowing
        operations.
    array_size: int
        Size of the original array (length of the time series) in which the
        sliding window operation will be applied.

    Returns
    -------
    pre_padding_size : int
    post_padding_size : int
    hum_hops : int
    """

    if (hop_size > window_size):
        raise ValueError("the `hop_size` must be <= than `window_size`")
    elif window_size < 2:
        raise ValueError("`the `window_size` must be >= than 2")

    half_window_size = window_size // 2

    pre_padding_size = half_window_size
    num_hops = (pre_padding_size + array_size - 1) // hop_size + 1
    remainder_samples = (pre_padding_size + array_size - 1) % hop_size

    # calculate the size of the post-padding
    post_padding_size = window_size - remainder_samples - 1

    return pre_padding_size, post_padding_size, num_hops



"""
Methods for playing audio.
"""
import sounddevice as sd
import numpy as np

import iracema.timeseries


def play(audio_time_series, blocking=False):
    """
    Play audio time series.

    Args
    ----
    audio_time_series: iracema.audio.Audio
    blocking: bool
    """
    sd.default.blocksize = 256
    sd.play(audio_time_series.data, audio_time_series.fs, blocking=blocking)


def play_with_clicks(audio_time_series, points, blocking=False):
    """
    Play audio with clicks in the instants corresponding to the specified
    points.

    Args
    ----
    audio_time_series: iracema.timeseries.Audio
    points: iracema.segments.PointList
    blocking: bool
    """
    click_sound = iracema.timeseries.Audio('/home/tairone/Development/iracema/audio/Click.wav')
    indexes = points.map_indexes(audio_time_series)
    audio_with_clicks = audio_time_series.copy()
    for i in indexes:
        audio_with_clicks.data[i:i+click_sound.nsamples] += click_sound.data

    sd.default.blocksize = 256
    sd.play(audio_with_clicks.data, audio_with_clicks.fs, blocking=blocking)

    return audio_with_clicks


def play_interval_samples(audio_time_series, from_sample, to_sample,
                          soft_start=True, blocking=False):
    """
    Play audio interval from sample `from_sample` to `to_sample`.

    Args
    ----
    audio_time_series: iracema.timeseries.Audio
    from_sample: int
    to_sample: int
    soft_start: bool
    blocking: bool
    """
    data_to_play = audio_time_series.data[from_sample:to_sample]
    size_soft_start = int(audio_time_series.fs * 0.005)
    soft_start = np.concatenate(
        (np.linspace(0, 1, size_soft_start, dtype=np.float_),
         np.ones(data_to_play.shape[-1] - size_soft_start)))

    sd.default.blocksize = 256
    sd.play(soft_start * data_to_play, audio_time_series.fs, blocking=blocking)


def play_interval_seconds(audio_time_series, from_seconds, to_seconds,
                          soft_start=True, blocking=False):
    """
    Play audio interval from time `from_seconds` to `to_seconds`.

    Args
    ----
    audio_time_series: iracema.timeseries.Audio
    from_seconds: int
    to_seconds: int
    soft_start: bool
    blocking: bool
    """
    if from_seconds:
        from_sample = int((from_seconds - audio_time_series.start_time) * audio_time_series.fs)
    else:
        from_sample = None
    if to_seconds:
        to_sample = int((to_seconds - audio_time_series.start_time) * audio_time_series.fs)
    else:
        to_sample = None

    play_interval_samples(audio_time_series, from_sample, to_sample,
                          soft_start=soft_start, blocking=blocking)


def stop():
    """
    Stop playing audio.
    """
    sd.stop()



"""
Functions that are commonly used in digital signal processing.
"""

import numpy as np


def local_peaks(array):
    """
    Find the local peaks of the `array`.

    Args
    ----
    array : numpy array
        Input array.

    Return
    ------
    values : numpy array
        Values for the local maxima.
    ix : numpy array
        Indexes of the local maxima in the input array.
    """
    # calculate the first derivative
    arr_diff = np.diff(array)
    # search for indexes where the slope changes from positive to negative
    ix = np.nonzero((arr_diff[1:] < 0) & (arr_diff[:-1] >= 0))[0] + 1
    values = array[ix]

    return values, ix


def n_highest_peaks(array, n):
    """
    Find the `n` highest peaks in the given `array`.

    Args
    ----
    array : numpy array
        Input array.
    n : int
        Number of peaks to search.
    """
    # sort the array by its peaks and choose the n highest values
    val, ix = local_peaks(array)
    ix_n = np.argsort(val)[-n:]
    # TODO: to figure out a more efficient way to to this

    return val[ix_n], ix[ix_n]


def hwr(array):
    """
    Half-wave rectifier
    """
    return np.clip(array, 0, None)


def decimate_mean(array, f):
    """
    Decimate array by grouping each `f` samples and taking their mean. The array
    will be padded with zeros at the end if its length is not divisible by `f`.
    """
    if array.ndim > 1:
        raise ValueError("array must be unidimensional")

    # pad the array with zeros at the end before reshaping
    pad = np.zeros((f - array.size % f,), dtype=array.dtype)
    padded_array = np.concatenate((array, pad))
    new_shape = (f, int(padded_array.size/f))
    reshaped_array = np.reshape(padded_array, new_shape)
    decimated_array = np.mean(reshaped_array, axis=0)

    return decimated_array



"""
Methods for reading and writing raw audio files.
"""
import numpy as np
import audioread
import scipy.io.wavfile


def read(filename):
    """
    Read WAV audio file.

    Arguments
    ---------
    filename: str
        Name of the audio file to be loaded.

    Return
    ------
    data: numpy array
        Audio data.
    fs: int
        Sampling frequency.
    """

    intmaxabs = 32768.  # maximum value for 16-bit signed integers
    fs = None

    with audioread.audio_open(filename) as input_file:
        fs = input_file.samplerate
        channels = input_file.channels

        # Audioread returns buffers containing 16-bit signed integers.
        data_int = np.array([], dtype=np.dtype('int16'))

        for frame in input_file:
            frame_int = np.frombuffer(frame, np.dtype('int16'))
            data_int = np.concatenate((data_int, frame_int), axis=0)

        # Convert data to float
        data = data_int.astype(
            np.float_, casting='safe')  # pylint: disable=maybe-no-member

        # Conversion to mono (mix both channels)
        if channels > 1:
            data = data.reshape((-1, channels)).T
            data = np.mean(data, axis=0)

        data = data / intmaxabs

    return (data, fs)


def write(filename, data, fs):
    """
    Write the contents of a `data` array to a WAVE file.

    Arguments
    ---------
    filename: str
        Name of the audio file to be loaded.
    data: numpy array
        Data array containing audio samples.
    fs: int
        Sampling frequency.
    """
    scipy.io.wavfile.write(filename, fs, data)



"""
Extraction of spectral information.
"""
import numpy as np
from iracema.aggregation import sliding_window


def spectrogram():
    """
    Generate spectrogram for time series.
    """
    pass


def fft(time_series, window_size, hop_size, fft_len=4096):
    """
    Calculate the Fast Fourier Transform for the ``time_series``

    Args
    ----
    time_series : TimeSeries
        Time series to be used in the FFT.
    window_size : int
    hop_size : int
    fftlen : int
        Length of the FFT. The signal will be zero-padded if ``fftlen`` >
        ``rolling_window.window_size``.
    """
    def calculate(x):
        return np.fft.fft(x, n=2 * fft_len)[:int(fft_len)]

    fft_time_series = sliding_window(time_series, window_size,
                                     hop_size, calculate, 'hann')

    fft_time_series.max_frequency = time_series.nyquist
    fft_time_series.frequencies = np.linspace(
        0, fft_time_series.max_frequency, fft_len, dtype=np.float_)

    fft_time_series.label = 'FFT'
    fft_time_series.unit = 'Magnitude'

    return fft_time_series


def ifft():
    """
    Inverse Fourier Transform.
    """
    pass



"""
Data plotting utilities.
"""
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import MultiCursor  # pylint: disable=import-error
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=import-error

import iracema.features as features


def plot_spectrogram(fft,
                     logfft=False,
                     fftlim=()):
    """
    Plot the spectrogram of the audio signal.
    """
    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))

    ax = f.subplots(1)

    plt.subplots_adjust(hspace=0.05)

    # plotting spectrogram
    add_spectrogram_to_axes(ax, fft, logfft, fftlim)

    # show the resulting image
    f.show()

    return f


def plot_audio_spectrogram(audio,
                           fft,
                           logfft=False,
                           rms=None,
                           peak_envelope=None,
                           fftlim=()):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal.
    """
    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    add_waveform_trio_to_axes(ax1, audio, rms, peak_envelope)

    # plotting spectrogram
    add_spectrogram_to_axes(ax2, fft, logfft, fftlim)

    # show the resulting image
    f.show()

    return f


def plot_audio_spectrogram_f0(audio,
                              rms,
                              peak_envelope,
                              fft,
                              fzero,
                              logfft=False,
                              fftlim=()):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal and its fundamental frequency  `fzero`.
    """

    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    add_waveform_trio_to_axes(ax1, audio, rms, peak_envelope)

    # plotting spectrogram
    add_spectrogram_to_axes(ax2, fft, logfft, fftlim)

    # plotting fzero
    add_curve_to_axes(ax2, fzero, fmt='r')

    # show the resulting image
    f.show()

    return f


def plot_audio_spectrogram_harmonics(audio,
                                     rms,
                                     peak_envelope,
                                     fft,
                                     fzero,
                                     harmonics,
                                     logfft=False,
                                     fftlim=()):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal and its `harmonics`.
    """

    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    add_waveform_trio_to_axes(ax1, audio, rms, peak_envelope)

    # plotting spectrogram
    add_spectrogram_to_axes(ax2, fft, logfft, fftlim)

    # plotting fzero
    add_curve_to_axes(ax2, fzero)

    # plotting harmonics
    add_curve_to_axes(ax2, harmonics)

    # show the resulting image
    f.show()

    return f


def plot_waveform_and_notes(audio, notes):
    """
    Plot waveform and note segments.
    """
    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))
    axes = f.subplots(1)

    add_waveform_to_axes(axes, audio)
    add_notes_to_axes(axes, notes)

    f.show()


def plot_waveform_trio(audio, rms=None, peak_envelope=None):
    """
    Plot a graph showing curves for the ``audio`` waveform, the ``rms`` and the
    ``peak_envelope``.
    """
    # configuring figure and subplots
    plt.figure(figsize=(15, 9))

    axes = plt.axes()

    # add waveform trio to first axes
    add_waveform_trio_to_axes(axes, audio, rms, peak_envelope)

    plt.show()


def plot_waveform_trio_and_features(audio,
                                       rms=None,
                                       peak_envelope=None,
                                       features=()):
    """
    Plot a graph showing curves for the ``audio`` waveform, the ``rms`` and the
    ``peak_envelope``; followed by a series of graphs, one for each time-series
    in the tuple `features`.
    """
    if not features:
        raise ValueError("the features to be plotted were not specified")

    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))

    axes_list = f.subplots(len(features) + 1, sharex=True)

    plt.subplots_adjust(hspace=0.05)

    # add audio to first axes
    add_waveform_trio_to_axes(axes_list[0], audio, rms, peak_envelope)

    # add features to the other axes
    for i, feature in enumerate(features, start=1):
        add_curve_to_axes(
            axes_list[i], feature, label=feature.label)
        axes_list[i].legend(loc='lower right', fontsize='x-small')

    MultiCursor(f.canvas, axes_list, color='gray', lw=1)

    plt.show()

    return f


def plot_waveform_trio_features_and_points(audio,
                                       feature,
                                       point_list,
                                       rms=None,
                                       peak_envelope=None):
    # configuring figure and subplots
    f = plt.figure(figsize=(15, 9))
    axes_list = f.subplots(2, sharex=True)

    # add waveform trio to first axes
    add_waveform_trio_to_axes(axes_list[0], audio, rms, peak_envelope)

    # add feature
    add_curve_to_axes(axes_list[1], feature, label=feature.label)
    axes_list[1].legend(loc='lower right', fontsize='x-small')

    add_points_to_axes(axes_list[1], point_list.time, point_list.get_values(feature))

    MultiCursor(f.canvas, axes_list, color='gray', lw=1)

    f.show()


def plot_spectrogram_3d(fft, logfft=False, fftlim=None):
    """
    Plot a 3D spectrogram (experimental feature).
    """
    # handle arguments
    if not fftlim:
        fmin, fmax = 20, fft.max_frequency
    else:
        fmin, fmax = fftlim[0], fftlim[1]
        if fmax > fft.max_frequency:
            warnings.warn(
                "the fmax specified excceeds the maximum frequency of the FFT")
            fmax = fft.max_frequency

    # configuring figure
    f = plt.figure(figsize=(15, 9))

    ax = Axes3D(f)

    # plotting spectrogram
    lenfft = fft.nfeatures
    bins_per_hz = lenfft / fft.max_frequency
    imax, imin = int(fmax * bins_per_hz), int(fmin * bins_per_hz)

    X, Y = np.meshgrid(fft.time, np.arange(imin, imax))

    fft_abs = (abs(fft.data[imin:imax, :]))

    ax.plot_surface(X, Y, fft_abs**2, rstride=2, cstride=10, cmap='seismic')

    if logfft:
        ax.set_yscale("log", basey=2)
        ax.get_yaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())

    # ax2.set(xlabel='time (s)', ylabel='frequency (Hz)')

    f.show()

    return f


def add_notes_to_axes(axes, notes):
    """
    Add note segments to the given ``axes``.
    """
    for note in notes:
        axes.plot(note.start_time, note.data[0], '^r')
        axes.plot(note.end_time, note.data[-1], 'vr')


def add_curve_to_axes(axes,
                      time_series,
                      fmt='b',
                      linewidth=0.8,
                      alpha=0.9,
                      label=None,
                      set_labels=True):
    "Add the curve for the given ``time_series`` to the given ``axes``."

    if set_labels:
        axes.set(ylabel=time_series.unit)
        if time_series.caption:
            axes.set_title(time_series.caption)

    axes.set_xlim(time_series.start_time, time_series.end_time)

    if time_series.nfeatures == 1:
        axes.plot(
            time_series.time,
            time_series.data,
            fmt,
            linewidth=linewidth,
            alpha=alpha,
            label=label)

    else:
        for row in time_series.data:
            axes.plot(
                time_series.time, row, fmt, linewidth=linewidth, alpha=alpha)


def add_points_to_axes(axes, time, values):
    """
    Plot circle in the specified ``points``x``values`` for the ``axes``.
    """
    axes.plot(time, values, 'or')


def add_waveform_to_axes(axes, audio):
    """
    Add the waveform for the time-series `audio` to `axes`.
    """

    # adding the curves
    add_curve_to_axes(axes, audio, linewidth=0.1, alpha=0.9)


def add_waveform_trio_to_axes(axes,
                              audio,
                              rms=None,
                              peak_envelope=None,
                              show_legend=True):
    """
    The waveform trio consists in the curves ``audio``, ``rms`` and
    ``peak_evelope``. This method adds them to ``axes``.
    """
    window_size, hop_size = 2048, 512
    rms = rms or features.rms(audio, window_size, hop_size)
    peak_envelope =\
        peak_envelope or features.peak_envelope(audio, window_size,
                                                   hop_size)
    # adding the curves
    add_curve_to_axes(axes, audio, linewidth=0.1, alpha=0.9)
    add_curve_to_axes(axes, rms, fmt='r', label=rms.label, set_labels=False)
    add_curve_to_axes(axes, peak_envelope, fmt='k', label=peak_envelope.label,
                      set_labels=False)

    if show_legend:
        axes.legend(loc='lower right', ncol=2, fontsize='x-small')


def add_segments_to_axes(axes, segments, start_color='g', end_color='r'):
    "Add two vertical lines to the axes delimiting a segment extent."


def add_spectrogram_to_axes(axes, fft, logfft=False, fftlim=()):
    """
    Add a spectrogram image to the axes
    """

    # handling arguments
    if not fftlim:
        fmin, fmax = 20, fft.max_frequency
    else:
        fmin, fmax = fftlim[0], fftlim[1]
        if fmax > fft.max_frequency:
            warnings.warn(
                "the fmax specified excceeds the maximum frequency of the FFT")
            fmax = fft.max_frequency

    # plotting spectrogram
    lenfft = fft.nfeatures
    bins_per_hz = lenfft / fft.max_frequency
    imax, imin = int(fmax * bins_per_hz), int(fmin * bins_per_hz)

    fft_abs = abs(fft.data[imin:imax, :])

    axes.imshow(
        20 * np.log10(fft_abs),
        origin='lower',
        aspect='auto',
        extent=[fft.start_time, fft.end_time, fmin, fmax],
        cmap='viridis')

    if logfft:
        axes.set_yscale("log", basey=2)
        axes.get_yaxis().set_major_formatter(
            matplotlib.ticker.ScalarFormatter())

    axes.set(xlabel='time (s)', ylabel='frequency (Hz)')



"""
According to the ANSI standard 1994 [ANSI1994]_, "Pitch is that attribute of
auditory sensation in terms of which sounds may be ordered on a scale extending
from low to high. Pitch depends mainly on the frequency content of the sound
stimulus, but it also depends on the sound pressure and the waveform of the
stimulus."

This module contains the implementation of a few different pitch estimation
methods.

References
----------
.. [ANSI1994] American National Standards Institute, 1994.
"""

import numpy as np
import scipy.signal as sig

from .timeseries import TimeSeries
from .util.dsp import local_peaks, n_highest_peaks, decimate_mean
from .aggregation import aggregate_features


def hps(fft_time_series, minf0, maxf0, n_downsampling=16,
        decimation='discard'):
    """
    Extract the pitch using Harmonic Product Spectrum.

    The Harmonic Product Spectrum measures the maximum coincidence for
    harmonics [Cuadra2001]_. It is based on successive downsampling operations
    on the frequency spectrum of the signal. If the signal contains harmonic
    components, then it should contain energy in the frequency positions
    corresponding to the integer multiples of the fundamental frequency. So by
    down-sampling the spectrum by increasing integer factors
    :math:`(1,2,3,...,R)` it is possible to align the energy of its harmonic
    components with the fundamental frequency of the signal.

    Then we multiply the original spectrum and its downsampled versions. This
    operation will make a strong peak appear in a position that corresponds to
    the fundamental frequency. The HPS calculates the maximum coincidence for
    harmonics, according to the equation:

    .. math:: Y(\\omega) = \\prod_{r=1}^{R} |X(\\omega r)|

    where :math:`X(\omega r)` represents one spectral frame and :math:`R` is
    the number of harmonics to be considered in the calculation. After this
    calculation a simple peak detection algorithm is used to obtain the
    fundamental frequency of the frame.

    This implementation modifies this approach adding an offset of 1 to the
    magnitude spectrum of the signal before applying the product shown in the
    equation above. This makes the algorithm more reliable in situations where
    some harmonics have very little or no energy at all (float arithmetic is
    not reliable when values get too close to zero).

    Also, alternatively to the original approach, it is possible to choose
    between different interpolation methods, using the argument `decimation`.

    Args
    ----
    fft_time_series : iracema.spectral.FFTs
        FFT time series.
    minf0 : float
        Lower frequency limit to search for f0.
    maxf0 : float
        Upper frequency limit to search for f0.
    n_downsampling : int
        Number of downsampling operations
    decimation : 'discard', 'mean' or 'interpolation'
        Type of decimation operation to be performed.

    References
    ----------
    .. [Cuadra2001] De La Cuadra, P. Efficient pitch detection techniques for
       interactive music. In ICMC, page 403–406, 2001.
    """
    if minf0 >= maxf0:
        raise ValueError('The parameter maxf0 must be greater than minf0.')

    N = fft_time_series.data.shape[-2]
    ix_maxf0 = np.int(np.ceil(
        (N * maxf0) /
        fft_time_series.max_frequency))  # max index to search for f0
    ix_minf0 = np.int(np.ceil(
        (N * minf0) /
        fft_time_series.max_frequency))  # min index to search for f0

    def calculate(X):
        """
        Calculate pitch for frame.
        """
        N = X.shape[-1]  # length of the FFT
        S = np.zeros((n_downsampling + 1, N),
                     dtype=np.float_)  # matrix with the decimated FFTs
        X_mag = np.abs(X)
        S[0, :] = X_mag
        for q in range(2, n_downsampling + 2):
            if decimation == 'mean':
                ds = decimate_mean(X_mag, q)
            elif decimation == 'interpolation':
                ds = sig.decimate(X_mag, q)
            elif decimation == 'discard':
                ds = X_mag[::q]
            S[q - 1, :ds.shape[-1]] = ds / q

        S = S + 1
        p_fzero = np.sum(S, axis=0)
        ix_f0 = ix_minf0 + np.argmax(p_fzero[ix_minf0:ix_maxf0])

        return fft_time_series.frequencies[ix_f0]

    pitch_time_series = aggregate_features(
        fft_time_series, calculate)
    pitch_time_series.label = 'Pitch (HPS)'
    pitch_time_series.unit = 'Hz'

    return pitch_time_series


def expan_pitch(fft_time_series,
                minf0=24,
                maxf0=4200,
                nharm=12,
                ncand=5,
                min_mag_cand=0.1,
                noisiness_tresh=0.99,
                perc_tol=0.04):
    """
    Extract the pitch using the Expan pitch detection algorithm.

    Args
    ----
    fft_time_series : iracema.spectral.FFTs
        FFT time series.
    minf0 : float
        Lower frequency limit to search for f0.
    maxf0 : float
        Upper frequency limit to search for f0.
    nharm : int
        Number of harmonics to be considered.
    ncand : int
        Number of f0 candidate components to be used.
    min_mag_cand : float
        Minimum magnitude of the candidate to be chosen as f0.
    noisiness_tresh : float
        Noisiness treshold.
    perc_tol : float
        Tolerance percentage to search for harmonics.
    """
    if minf0 >= maxf0:
        raise ValueError('The parameter maxf0 must be greater than minf0.')
    if nharm < 3:
        raise ValueError('The parameter nharm must be greater than 3.')

    N = fft_time_series.data.shape[-2]
    ix_maxf0 = np.int(np.ceil(
        (N * maxf0) /
        fft_time_series.max_frequency))  # max index to search for f0
    ix_minf0 = np.int(np.ceil(
        (N * minf0) /
        fft_time_series.max_frequency))  # min index to search for f0

    def frame_pitch(fft_frame):
        """
        Extract the pitch and the harmonics for a fft frame.
        """
        N = fft_frame.shape[-1]  # length of the FFTT
        fft_frame_mag = np.abs(fft_frame)

        # searching for f0 candidates
        _, ix_cand = n_highest_peaks(fft_frame_mag[ix_minf0:ix_maxf0], ncand)
        ix_cand = ix_cand + ix_minf0  # TODO: check these indexes

        # searching for local peaks all over the FFT
        _, ix_pks = local_peaks(fft_frame_mag[ix_minf0:])
        ix_pks = ix_pks + ix_minf0
        # TODO: this could probably be done more efficiently, the peak searching
        # is being performed twice

        # update the number of candidates with the number of local peaks found
        n_cand = len(ix_cand)

        # if no peaks were found, return zeroed arrays
        if not ix_cand.any():
            return 0

        # memory allocation
        cand_mag = np.zeros((ncand, nharm))
        ix_cand_harm = np.zeros((ncand, nharm))

        # iterate the candidates
        for i, ix_f0 in enumerate(ix_cand):
            delta = ix_f0 * perc_tol
            cand_mag[i, 0] = fft_frame_mag[ix_f0]

            # test the candidate's magnitude compared to the frame's max
            if cand_mag[i, 0] > min_mag_cand * np.max(fft_frame_mag):
                ix_cand_harm[i, 0] = ix_f0

                # iterate the candidate's potential harmonics
                for j in range(1, nharm):
                    ix_f_harm = ix_f0 * (j + 1)

                    # check if it is within the boundaries of the FFT
                    if ix_f_harm < N:
                        # set the tolerance margin to search for the harmonic
                        ix_tol_low = int(ix_f_harm - delta)
                        ix_tol_hi = int(ix_f_harm + delta)
                        if ix_tol_low < ix_minf0:
                            ix_tol_low = ix_minf0

                        # search for highest peak within the tolerance margin
                        val_harm, ix_harm = n_highest_peaks(
                            fft_frame_mag[ix_tol_low:ix_tol_hi], 1)
                        ix_harm = ix_harm + ix_tol_low

                        if ix_harm:
                            cand_mag[i, j] = val_harm
                            ix_cand_harm[i, j] = ix_harm

                        else:
                            cand_mag[i, j] = fft_frame_mag[(j + 1) * ix_f0]
                            ix_cand_harm[i, j] = (j + 1) * ix_f0
                    else:
                        # stop the loop if the frequency index is out of bound
                        break

        energy_harm = np.zeros(n_cand)

        # calculate the energy of the harmonic components for each candidate
        for i in range(n_cand):
            energy_harm[i] = np.sum(cand_mag[i, :]**2)

        # choose the candidate with the highest harmonic energy
        i = np.argmax(energy_harm)

        # one last test, the noisiness for the winner candidate must be bellow
        # the noisiness threshold
        h_energy = energy_harm[i]
        frame_energy = np.sum(fft_frame_mag**2)
        frame_noisiness = 1 - (h_energy / frame_energy)

        if frame_noisiness < noisiness_tresh:
            return ix_cand_harm[i, 0] * fft_time_series.max_frequency / N
        else:
            return 0

    pitch_time_series = aggregate_features(
        fft_time_series, frame_pitch)

    pitch_time_series.label = 'Pitch (HPS)'
    pitch_time_series.unit = 'Hz'


    return pitch_time_series


def crepe_pitch(audio,
                model_capacity='large',
                min_confidence=0.5,
                viterbi=True):
    """
    Extract the pitch using CREPE pitch tracker.

    This function uses a pitch tracker based on deep convolutional neural
    networks. The model was proposed and trained by [Kim2018]_.

    Args
    ----
    audio : iracema.audio.Audio
        Audio time series.
    step_size : float
        Length of the time steps for the pitch extraction.
    model_capacity : 'tiny', 'small', 'medium', 'large', or 'full'
        String specifying the capacity of the model. The value 'full' will use
        the model specified in the paper. The others will reduce the number of
        filters in the convolutional layers, resulting in faster computation,
        at the cost of slightly reduced accuracy.
    min_confidence : float
        Minimum confidence to consider a pitch detection as valid.
    viterbi : bool
        Viterbi smoothing for pitch curve.

    References
    ----------
    .. [Kim2018] Kim, J. W., Salamon, J., Li, P., & Bello, J. P. (2018). CREPE:
       A Convolutional Representation for Pitch Estimation. In Proceedings of
       the IEEE International Conference on Acoustics, Speech, and Signal
       Processing (ICASSP).
    """
    from crepe import predict  # pylint: disable=import-error

    time, frequency, confidence, _ = predict(
        audio.data, audio.fs, viterbi=viterbi, model_capacity=model_capacity)

    frequency[confidence < min_confidence] = np.nan

    step = time[1] - time[0]
    print('step =', step)
    fs = 1 / step
    pitch_time_series = TimeSeries(
        fs, frequency, start_time=audio.start_time, unit='Hz')

    return pitch_time_series


def pitch_filter(pitch_time_series, delta_max=0.04):
    """
    The pitch curve can be noisy, this function tries to smooth it.
    """
    data = pitch_time_series.data
    data_previous = np.concatenate((np.zeros(1), data[0:-1]))
    data_next = np.concatenate((data[1:], np.zeros(1)))

    pitch_filtered = pitch_time_series.copy()

    # find the indexes where the pitch curve is unstable
    # 1st condition: the pitch changes more than delta_max and comes back
    diff_next_prev = np.abs(data_next - data_previous)
    diff_data_prev = np.abs(data - data_previous)
    diff_data_next = np.abs(data - data_next)
    mean_next_prev = (data_next + data_previous) / 2
    indexes_to_interp = ((diff_next_prev < mean_next_prev * delta_max) &
                         ((diff_data_prev > data * delta_max) |
                          (diff_data_next > data * delta_max)))

    for idx in np.where(indexes_to_interp):
        pitch_filtered.data[idx] = (
            pitch_filtered.data[idx + 1] + pitch_filtered.data[idx - 1]) / 2

    # 2nd condition: one isolated point between zeros
    indexes_to_zero = ((data_previous == 0) & (data_next == 0))

    pitch_filtered.data[indexes_to_zero] = 0

    return pitch_filtered



"""
iracema is a python package aimed at the extraction of expressive music
information from audio signals
"""

from .timeseries import TimeSeries, Audio
from .segment import Segment, SegmentList

import iracema.features
import iracema.harmonics
import iracema.pitch
import iracema.plot
import iracema.segmentation
import iracema.spectral



"""
A couple of different methods for note segmentation.
"""

import numpy as np
import scipy.signal as sig

import iracema.features
import iracema.pitch
import iracema.segment
from iracema.aggregation import aggregate_sucessive_samples

from iracema.plot import plot_waveform_trio_features_and_points


def extract_note_onsets_rms(audio, rms=None, min_time=None,
                            perc_threshold_pk=0.2):
    """
    Extract note onsets from the ``audio`` time-series using its ``rms``.
    The RMS will be calculated if it's not passed as an argument. The argument
    ``min_time`` can be used to specify the minimum distance (in seconds)
    between two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    rms : iracema.TimeSeries, optional
        Pre-calculated RMS for the audio time-series.
    min_time : float, optional
        Minimum time (in seconds) between successive onsets.
    perc_threshold_pk : float
        A percentual of the ODF maximum to be defined as a threshold
        for the peak picking.

    Return
    ------
    onsets : list
        List of onset points.
    """
    rms = rms or iracema.features.rms(audio, 2048, 512)
    # TODO: hardcoded parameters should be obtained in a better way

    # handling arguments
    if min_time:
        min_dist = int(min_time * pitch.fs)
        if min_dist == 0:
            min_dist = None
    else:
        min_dist = None

    # onset detection function
    onset_df = detection_function_rms(rms)

    # peak picking
    threshold = perc_threshold_pk * np.max(onset_df.data)
    ix_onsets, _ = sig.find_peaks(
        onset_df.data, height=threshold, distance=min_dist)

    print(ix_onsets)

    onsets = iracema.segment.PointList([
        iracema.segment.Point(rms, position)
        for position in ix_onsets
    ])

    print(onsets.time, onsets.get_values(onset_df))

    plot_waveform_trio_features_and_points(audio, onset_df, onsets)

    return onsets


def extract_note_onsets_pitch(audio, pitch, min_time=None, delta_pitch_ratio=0.03):
    """
    Extract note onsets from the ``audio`` time-series using its ``pitch``.
    The argument ``min_time`` can be used to specify the minimum distance (in
    seconds) between two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    pitch : iracema.TimeSeries
        Pre-calculated pitch for the audio time-series.
    min_time : float, optional
        Minimum time (in seconds) between successive onsets.

    Return
    ------
    onsets : list
        List of onset points.
    """

    # handling arguments
    if min_time:
        min_dist = int(min_time * pitch.fs)
        if min_dist == 0:
            min_dist = None
    else:
        min_dist = None

    # onset detection function
    onset_df = detection_function_pitch(pitch)

    # peak picking
    ix_onsets, _ = sig.find_peaks(
        onset_df.data, height=delta_pitch_ratio, distance=min_dist)

    onsets = iracema.segment.PointList([
        iracema.segment.Point(pitch, position)
        for position in ix_onsets
    ])

    #plot_waveform_trio_features_and_points(audio, onset_df, onsets)

    return onsets


def segment_notes_rms(audio, rms=None, min_time=None, perc_threshold_pk=0.05):
    """
    Extract note segments from the ``audio`` time-series using its ``rms``.
    The RMS will be calculated if it's not passed as an argument. The argument
    ``min_time`` can be used to specify the minimum distance (in seconds)
    between two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    rms : iracema.TimeSeries, optional
        Pre-calculated RMS for the audio time-series.
    min_time : float, optional
        Minimum time (in seconds) between successive onsets.
    perc_threshold_pk : float
        A percentual of the ODF maximum to be defined as a threshold
        for the peak picking.

    Return
    ------
    notes : list
        List of `Note` segments.
    """

    # handling arguments
    if min_time:
        min_dist = int(min_time * pitch.fs)
        if min_dist == 0:
            min_dist = None
    else:
        min_dist = None

    rms = rms or iracema.features.rms(audio, 2048, 512)
    # TODO: hardcoded parameters should be obtained in a better way

    # onset detection function
    onset_df = detection_function_rms(rms)

    # peak picking
    threshold = perc_threshold_pk * np.max(onset_df.data)
    ix_onsets, _ = sig.find_peaks(
        onset_df.data, height=threshold, distance=min_dist)

    # map the indexes to the original time-series
    ix_onsets_original = onset_df.map_index_to_original(ix_onsets)

    # offset detection
    ix_offsets = np.empty_like(ix_onsets)
    ix_offsets[-1] = onset_df.nsamples
    ix_offsets[:-1] = ix_onsets[1:] - 1

    # map the indexes to the original time-series
    ix_offsets_original = onset_df.map_index_to_original(ix_offsets)

    notes = [
        Segment(audio, st, end)
        for st, end in zip(ix_onsets_original, ix_offsets_original)
    ]

    return notes


def segment_notes_pitch(audio, pitch, min_time=None):
    """
    Extract note segments from the ``audio`` time-series using its ``pitch``. The
    pitch will be calculated if it's not passed as an argument. The argument
    `min_time` can be used to speficy the minimum distance (in seconds) between
    two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    pitch : TimeSeries, optional
        Pitch calculated from the audio time-series.
    min_time : float, optional
        Minimum time (in seconds) between successive onsets.

    Return
    ------
    notes : list
        List of `Note` segments.

    """
    if min_time:
        min_dist = int(np.ceil(min_time * pitch.fs))
    else:
        min_dist = None

    # onset detection function
    onset_df = detection_function_pitch(pitch)

    ix_onsets, _ = sig.find_peaks(onset_df.data, distance=min_dist)

    # map the indexes to the original time-series
    ix_onsets_original = onset_df.map_index_to_original(ix_onsets)

    # peak picking


def get_notes_list(audio, onsets, offsets):
    """
    Generate a list of note segments using the specified `onsets` and `offsets`
    arrays.

    Args
    ----
    audio : Audio
        Audio object
    onsets : array
        Indexes of the onset occurrences in `audio`.
    offsets : array
        Indexes of the offset occurrences in `audio`.

    Return
    ------
    notes : list
        List of segments.
    """
    if onsets.shape != offsets.shape:
        raise ValueError("the number of onsets and offsets must the same")

    return [Segment(audio, onsets[i], offsets[i]) for i in range(len(onsets))]


def detection_function_rms(rms):
    """
    Onset detection function based on RMS.
    """
    return rms.diff().hwr() #* rms


def detection_function_pitch(pitch):
    """
    Onset detection function based on Pitch.
    """
    def ratio_successive(current, last):
        return abs((current / last) - 1)

    return aggregate_sucessive_samples(pitch, ratio_successive)



"""
Contain classes used to extract and manipulate segments.
"""
import numpy as np

from .util import conversion


class Segment:
    """
    The objects generated from this class are used to retrieve excerpts from
    time-series.

    The `start` and `end` arguments can be specified in terms of sample index
    or time (in seconds). The argument `limits_unit` must be set accordingly.

    Args
    ----
    time_series : TimeSeries
        Original time series related to the segment.
    start : int or float
        Index (or sample number) corresponding to the start of the segment in
        the time-series from which it derived. Alternatively, this value can
        be specified in seconds.
    end : int or float
        Index of the ending sample for the segment. Alternatively, this value
        can be specified in seconds.
    limits_unit : ("sample_index", "seconds")
        If 'sample_index' is passed (default), the arguments `start` and `end`
        must be integers corresponding to sample indexes whitin `time_series`.
        Else, if 'seconds' is passed, both arguments must correspond to the
        time of these limits, in seconds.
    """
    def __init__(self, time_series, start, end, limits_unit='sample_index'):
        if limits_unit not in ('sample_index', 'seconds'):
            raise ValueError("invalid value for `limits_unit` argument: must" +
                             " be 'sample_index' or 'seconds'")

        if end is not None and start > end:
            raise ValueError("end must be > start")

        self.fs = time_series.fs
        self.time_offset = time_series.start_time

        if limits_unit == 'sample_index':
            if type(start) != int or type(end) != int:
                raise ValueError("`start` and `end` must be of type int when" +
                                 "`limits_unit`=='sample_index'")
            self.start = start
            self.end = end

        elif limits_unit == 'seconds':
            self.start = conversion.seconds_to_sample_index(start, self.fs)
            self.end = conversion.seconds_to_sample_index(end, self.fs)

    @property
    def nsamples(self):
        return self.end - self.start

    @property
    def start_time(self):
        return conversion.sample_index_to_seconds(self.start, self.fs,
                                                  self.time_offset)

    @property
    def end_time(self):
        return conversion.sample_index_to_seconds(self.end, self.fs,
                                                  self.time_offset)

    def generate_slice(self, time_series):
        """
        Generate a python slice with the sample indexes that correspond to the
        current segment in `time_series`
        """
        new_fs = time_series.fs
        new_time_offset = time_series.start_time

        if self.fs == time_series.fs:
            return slice(self.start, self.end)
        else:
            slice_start = conversion.map_sample_index(
                self.start, self.fs, self.time_offset, new_fs, new_time_offset)
            slice_end = conversion.map_sample_index(
                self.end, self.fs, self.time_offset, new_fs, new_time_offset)

            return slice(slice_start, slice_end)

    def __repr__(self):
        "Overload the representation for the class"
        class_name = self.__class__.__name__
        description = "(start: {}, end: {}), fs: {}".format(
            self.start, self.end, self.fs)
        return '{}: {}'.format(class_name, description)


class SegmentList(list):
    """
    List of segments.
    """

    def __init__(self, segment_list):
        super(SegmentList, self).__init__(segment_list)


class Point:
    """
    A point object represents an instant in a time series, i.e., one specific
    sample index. It is flexible enough to locate samples corresponding to the
    same instant in time series with different sampling rates.

    Args
    ----
    time_series : TimeSeries
        Original time series related to the point.
    position : int or float
        Index (or sample number) corresponding to the position of the point in
        the time-series from which it derived. Alternatively, this value can
        be specified in seconds.
    unit : ("sample_index", "seconds")
        If 'sample_index' is passed (default), the argument `position` must be
        an integer corresponding to a sample index whitin `time_series`. Else,
        if 'seconds' is passed, `position` must be specified in terms of time.
    """
    def __init__(self, time_series, position, unit='sample_index'):
        if unit not in ('sample_index', 'seconds'):
            raise ValueError("invalid value for `unit` argument: must" +
                             " be 'sample_index' or 'seconds'")

        self.fs = time_series.fs
        self.time_offset = time_series.start_time

        if unit == 'sample_index':
            if type(position) != np.int_:
                raise ValueError("`position` must be of type int when" +
                                 "`limits_unit`=='sample_index'")
            self.position = position

        elif unit == 'seconds':
            self.position = conversion.seconds_to_sample_index(
                position, self.fs)
        else:
            raise ValueError("`unit` must be 'sample_index' or 'seconds'")

    @property
    def time(self):
        return conversion.sample_index_to_seconds(
            self.position, self.fs, self.time_offset)

    def map_index(self, time_series):
        new_fs = time_series.fs
        new_time_offset = time_series.start_time
        return conversion.map_sample_index(
            self.position, self.fs, self.time_offset, new_fs, new_time_offset)

    def get_value(self, time_series):
        sample_index = self.map_index(time_series)
        return time_series.data[sample_index]


class PointList(list):
    """
    List of points.
    """
    def __init__(self, point_list):
        super(PointList, self).__init__(point_list)

    @property
    def time(self):
        return [point.time for point in self]

    def map_indexes(self, time_series):
        return [point.map_index(time_series) for point in self]

    def get_values(self, time_series):
        return [point.get_value(time_series) for point in self]



"""
This module contains the implementation of some classic feature extractors.

References
----------
.. [Bello2005] Bello, J. P., Daudet, L., Abdallah, S., Duxbury, C., Davies,
    M., & Sandler, M. B. (2005). A tutorial on onset detection in music
    signals. IEEE Transactions on Speech and Audio Processing, 13(5),
    1035–1046.

.. [Dixon2006] Dixon, S. (2006). Onset Detection Revisited. In 9th
   International Conference on Digital Audio Effects (pp. 133–137).
   Montreal, Canada.

.. [Lerch2012] Lerch, A. (2012). An introduction to audio content analysis:
   Applications in signal processing and music informatics. In An Introduction
   to Audio Content Analysis: Applications in Signal Processing and Music 
   Informatics.

.. [Park2004] Park, T. H. (2004). Towards automatic musical instrument
   timbre recognition. Princeton University.

.. [Park2010] Park, T. H. (2010). Introduction to digital signal
   processing: Computer musically speaking. World Scientific Publishing
   Co. Pte. Ltd.

.. [Peeters2011] Peeters, G., Giordano, B. L., Susini, P., Misdariis, N.,
   & McAdams, S. (2011). The timbre toolbox: extracting audio features
   from musical signals, 130(5).

"""

import numpy as np
from scipy.stats import gmean  # pylint: disable=import-error

from .aggregation import (aggregate_features, aggregate_sucessive_samples,
                          sliding_window)
from .util.dsp import hwr


def peak_envelope(time_series, window_size, hop_size):
    """
    Calculate the peak envelope of a time series

    The peak envelope consists in the peak absolute values of the
    amplitude within the aggregation window.

    .. math:: \\operatorname{PE} = max(|x(n)|), 1 <= n <= L

    Where `x(n)` is the n-th sample of a window of length `L`.

    Args
    ----
    time_series : iracema.timeseries.TimeSeries
        An audio time-series object.
    window_size : int
    hop_size : int
    """
    def function(x):
        return np.max(np.abs(x))

    time_series = sliding_window(time_series, window_size, hop_size,
                                 function)
    time_series.label = 'PeakEnvelope'
    time_series.unit = 'amplitude'
    return time_series


def rms(time_series, window_size, hop_size):
    """
    Calculate the root mean square of a time series

    The RMS envelope consists in the root mean square of the amplitude,
    calculated within the aggregation window.

    .. math:: RMS = \\sqrt{ \\frac{1}{L} \\sum_{n=1}^{L} x(n)^2 }

    Where `x(n)` is the n-th sample of a window of length `L`.

    Args
    ----
    time_series : iracema.timeseries.TimeSeries
        A time-series object. It is usually applied on Audio objects.
    window_size : int
    hop_size : int
    """
    def function(x):
        return np.sqrt(np.mean(x**2))

    time_series = sliding_window(time_series, window_size, hop_size,
                                 function)
    time_series.label = 'RMS'
    time_series.unit = 'amplitude'
    return time_series


def zcr(time_series, window_size, hop_size):
    """
    Calculate the zero-crossing rate of a time series, i.e., the number of
    times the signal crosses the zero axis, per second.

    The zero-crossing rate gives some insight on the noisiness character of a
    sound. In noisy / unvoiced signals, the zero-crossing rate tends to reach
    higher values than in periodic / voiced signals.

    .. math:: \\operatorname{ZC} = \\frac{1}{2 L} \\sum_{n=1}^{L}\\left|\\operatorname{sgn}\\left[x(n)\\right]-\\operatorname{sgn}\\left[x(n-1)\\right]\\right|

    Where

    .. math:: \\operatorname{sgn}\\left[x(n)\\right]=\\left\\{\\begin{array}{c}{1, x(n) \\geq 0} \\\\ {-1, x(n)<0}\\end{array}\\right.

    And `x(n)` is the n-th sample of a window of length `L`.

    Args
    ----
    time_series : iracema.timeseries.TimeSeries
        A time-series object. It is usually applied on Audio objects.
    window_size : int
    hop_size : int
    """
    # count the number of times the signal changes between successive samples
    def function(x):
        return np.sum(x[1:] * x[:-1] < 0) / window_size * time_series.fs

    time_series = sliding_window(time_series, window_size, hop_size,
                                 function)
    time_series.label = 'ZCR'
    time_series.unit = 'Hz'
    return time_series


def spectral_flatness(fft):
    """
    Calculate the spectral flatness for a given FFT.

    The spectral flatness gives an estimation of the noisiness / sinusoidality
    of an audio signal (for the whole spectrum or for a frequency range). It
    can be used to determine voiced / unvoiced parts of a signal [Park2004]_.

    It is defined as the ratio between the `geometric mean` and the
    `arithmetic mean` of the energy spectrum:

    .. math::
       :nowrap:

       \\begin{eqnarray}
       \\operatorname{SFM} = 10 log_{10} \\left( \\frac
         {\\left( \\prod_{k=1}^{N} |X(k)| \\right)^\\frac{1}{N}}
         { \\frac{1}{N} \\sum_{k=1}^{N} |X(k)| }
       \\right)
       \\end{eqnarray}

    Where `X(k)` is the result of the FFT for the `k-th` frequency bin.

    Args
    ----
    time_series : iracema.spectral.FFT
        A FFT object
    """
    def function(X):
        fft_magnitudes = np.abs(X)
        return 10 * np.log10(gmean(fft_magnitudes) / np.mean(fft_magnitudes))

    time_series = aggregate_features(fft, function)
    time_series.label = 'SpectralFlatness'
    time_series.unit = ''
    return time_series


def hfc(fft, method='energy'):
    """
    Calculate the high frequency content for a FFT time-series.

    The HFC _function produces sharp peaks during attacks or transients
    [Bello2005]_ and might be a good choice for detecting onsets in percussive
    sounds.

    .. math:: \\operatorname{HFC} = \sum_{k=1}^{N} |X(k)|^2 \\cdot k

    Alternatively, you can set ``method`` = `'amplitude'` instead of `'energy'`
    (default value):

    .. math:: \\operatorname{HFC} = \sum_{k=1}^{N} |X(k)| \\cdot k

    Args
    ----
    fft : iracema.spectral.FFT
        FFT time-series.
    method : str
        Method of choice to calculate the HFC.

    """
    def _func(X):
        N = X.shape[0]
        W = np.arange(1, N + 1)

        if method == 'energy':
            return np.sum(W * np.abs(X)**2) / N
        elif method == 'amplitude':
            return np.sum(W * np.abs(X)) / N
        else:
            ValueError("the argument `method` must be 'energy' or 'amplitude'")

    time_series = aggregate_features(fft, _func)
    time_series.label = 'HFC'
    time_series.unit = ''
    return time_series


def spectral_centroid(fft):
    """
    Calculate the spectral centroid for a FFT time-series.

    The spectral centroid is a well known timbral feature that is used to
    describe the brightness of a sound. It represents the center of gravity
    of the frequency components of a signal [Park2010]_.

    .. math::
       \\operatorname{SC} = \\frac{\\sum_{k=1}^{N} |X(k)| \\cdot f_k }{\\sum_{k=1}^{N} |X(k)|}

    Where `X(k)` is the result of the FFT for the `k-th` frequency bin.

    Args
    ----
    fft : iracema.spectral.FFT
        A FFT object

    """
    def function(X):
        return __spectral_centroid(X, fft.frequencies)

    time_series = aggregate_features(fft, function)
    time_series.label = 'SpectralCentroid'
    time_series.unit = 'Hz'
    return time_series


def spectral_spread(fft):
    """
    Calculate the spectral spread for a FFT time-series.

    The spectral spread represents the spread of the spectrum around the
    spectral centroid [Peeters2011]_, [Lerch2012]_.

    .. math:: \\operatorname{SSp} = \\sqrt{\\frac{\\sum_{k=1}^{N} |X(k)| \\cdot (f_k - SC)^2 }{\\sum_
       {k=1}^{N} |X (k)|}}

    Where `X(k)` is the result of the FFT for the `k-th` frequency bin and SC
    is the spectral centroid for the frame.


    """
    def function(X):
        return __spectral_spread(X, fft.frequencies)

    time_series = aggregate_features(fft, function)
    time_series.label = 'SpectralSpread'
    time_series.unit = 'Hz'
    return time_series


def __spectral_centroid(X, f):
    """
    Calculate the spectral centroid for a fft frame `X`, being `f` the
    frequency corresponding to its bins.
    """
    abs_X = np.abs(X)
    sum_abs_X = np.sum(abs_X)
    if sum_abs_X == 0:
        return 0
    return np.sum(f * abs_X) / sum_abs_X


def __spectral_spread(X, f):
    """
    Calculate the spectral spread for a fft frame `X`, being `f` the frequency
    corresponding to its bins.
    """
    return np.sqrt(__spectral_centroid(X, (f - __spectral_centroid(X, f))**2))



def spectral_skewness(fft):
    """
    Calculate the spectral skewness for an FFT time series
    
    The spectral skewness is a measure of the asymetry of the distribution of
    the spectrum around its mean value, and is calculated from its third order
    moment. It will output negative values when the spectrum has more energy
    bellow the mean value, and positive values when it has more energy above
    the mean. Symmetric distributions will output the value zero [Lerch2012]_.

    .. math::
       \\operatorname{SSk} = \\frac{2 \\cdot \\sum_{k=1}^{N} \\left( |X(k)| - \\mu_{|X|} \\right)^3 }{
       N \\cdot \\sigma_{|X|}^3}

    Where :math:`\\mu_{|X|}` is the mean value of the maginute spectrum and 
    :math:`\\sigma_{|X|}` its standard deviation.

    """

    def _func(X):
        pass


def spectral_kurtosis(fft):
    """
    Calculate the spectral kurtosis for an FFT time series
    
    The spectral kurtosis is a measure of the flatness of the distribution of
    the spectrum around its mean value. It will output the value 3 for Gaussian
    distributions. Values smaller than 3 represent flatter distributions, while
    values larger than 3 represent peakier distributions [Lerch2012]_.

    .. math::
       \\operatorname{SKu} = \\frac{2 \\cdot \\sum_{k=1}^{N} \\left( |X(k)| - \\mu_{|X|} \\right)^4 }{
       N \\cdot \\sigma_{|X|}^4}

    Where :math:`\\mu_{|X|}` is the mean value of the maginute spectrum and 
    :math:`\\sigma_{|X|}` its standard deviation.


    """
    def _func(X):
        pass


def spectral_flux(fft):
    """
    Calculate the spectral flux for a FFT time-series.

    The spectral flux measures the amount of change between successive
    spectral frames. There are different methods to calculate the spectral
    flux across the literature. For now we have implemented the one proposed
    by [Dixon2006]_.

    .. math:: \\operatorname{SF} = \\sum_{k=1}^{N} H(|X(t, k)| - |X(t-1, k)|)

    where :math:`H(x) = \\frac{x+|x|}{2}` is the half-wave rectifier _function,
    and `t` is the temporal index of the frame.

    Args
    ----
    fft : iracema.spectral.FFT
        A FFT object

    """
    def function(X, X_prev):
        return np.sum(hwr(np.abs(X) - np.abs(X_prev)))

    time_series = aggregate_sucessive_samples(fft, function)
    time_series.label = 'SpectralFlux'
    time_series.unit = ''
    return time_series


def spectral_rolloff(fft):
    """Spectral Rolloff"""
    def _func(X):
        pass


def spectral_irregularity(fft):
    """Spectral Irregularity"""
    def _func(X):
        pass


def harmonic_centroid(harmonics):
    """
    Harmonic Centroid

    The harmonic centroid represents the center of gravity of the amplitudes
    of the harmonic series.

    .. math::
       \\operatorname{HC} = \\frac{\\sum_{k=1}^{H} A(k) \\cdot f_k }{\\sum_{k=1}^{H} A(k)}

    Where :math:`A(h)` represents the amplitude of the h-th harmonic partial.
    """

    def _func(X):
        pass


def inharmonicity(fft, harmonics):
    """Inharmonicity"""
    def _func(X):
        pass


def harmonic_energy(harmonics_magnitude):
    """
    Calculate the energy of harmonic partials.

    Harmonic energy is the energy of the harmonic partials of a signal.

    .. math:: \\operatorname{HE} = \\sum_{k=1}^{H} A(k)^2

    """
    def function(frame):
        return np.sum(frame**2)

    time_series = aggregate_features(harmonics_magnitude, function)
    time_series.label = 'Harmonic Energy'
    time_series.unit = ''
    return time_series


def spectral_entropy(fft):
    """
    Calculate the spectral entropy for a FFT time series

    The spectral entropy is based on the concept of information entropy from
    Shannon's information theory. It measures the unpredictability of the given
    state of a spectral distribution.

    .. math:: \\operatorname{SEpy} = - \\sum_{k}^{N} P(k) \\cdot \\log_2 P(k)

    Where 

    .. math:: P(i)=\\frac{|X(i)|^2}{\sum_{j}^{N} |X(j)|^2}

    More info at https://www.mathworks.com/help/signal/ref/pentropy.html.
    """
    def function(X):
        N = fft.nfeatures
        P = np.abs(X)**2 / np.sum(np.abs(X)**2)
        H = -(np.sum(P * np.log2(P))) / np.log2(N)
        return H

    time_series = aggregate_features(fft, function)
    time_series.label = 'Spectral Entropy'
    time_series.unit = ''
    return time_series


def spectral_energy(fft):
    """
    Calculate the total energy of an FFT frame.

    Spectral Energy is the total energy of an FFT frame.

    .. math:: \\operatorname{SF} = \\sum_{k=1}^{N} H(|X(t, k)| - |X(t-1, k)|)
    """
    def function(frame):
        return np.sum(np.abs(frame)**2)

    time_series = aggregate_features(fft, function)
    time_series.label = 'Spectral Energy'
    time_series.unit = ''
    return time_series


def noisiness(fft, harmonics_magnitude):
    """
    Calculate the Noisiness for the given FFT and Harmonics time series.

    The Noisiness represent how noisy a signal is (values closer to 1), as
    oposed to harmonic (values close to 0). It is the ratio of the noise
    energy to the total energy of a signal [Peeters2011]_.

    .. math:: \\operatorname{Ns} = \\frac{\\operatorname{SE}-\\operatorname{HE}}{\\operatorname{SE}}

    """
    energy_spectral = spectral_energy(fft)
    energy_harmonic = harmonic_energy(harmonics_magnitude)
    energy_noise = energy_spectral - energy_harmonic

    time_series = energy_noise / energy_spectral
    time_series.label = 'Noisiness'
    time_series.unit = ''

    return time_series


def oer(harmonics):
    """
    Calculate the odd-to-even ratio for the harmonics time series.

    The OER represents the odd-to-even ratio among the harmonics of an audio
    signal. This value will be higher for sounds with predominantly odd
    harmonics, such as the clarinet.
    
    .. math:: \\operatorname{OER}=\\frac{\\sum_{h=1}^{H / 2} A(2 h - 1)^{2}\\left(t_{m}\\right)}{\\sum_{h=1}^{H / 2} A(2 h)^{2}\\left(t_{m}\\right)}

    Where :math:`A(h)` represents the amplitude of the h-th harmonic partial.
    """
    def _func(X):
        pass




"""
Implementation of time series.
"""

import copy as cp
from os import path

import matplotlib.pyplot as plt
import numpy as np

from .segment import Segment
from .util import conversion
from .io.audiofile import read
from .io import player


class TimeSeries:
    """
    Base class for time series objects, which can represent synchronous
    discrete time series.

    Args
    ----
    fs : float
        Sampling frequency for the data.
    data : numpy array, optional
        Data array sampled at ``fs`` Hz. If this argument is not provided,
        the method _write_data() must be called after the initialization to
        set the data array.
    start_time : float, optional
        The time in seconds the time series start, relative to the original
        time reference.

    Attributes
    ----------
    data : numpy array
        Data array sampled at ``fs`` Hz.
    time : numpy array
        Numpy data array containing the time of each sample, relative to the
        original time reference.
    fs : float
        Sampling frequency for the data.
    nyquist : float
        Nyquist frequency for the data.
    ts : float
        Sampling period for the data.
    start_time : float
        The time the time series start (in seconds) relative to the original
        time reference.
    duration : float
        Duration of the time series (in seconds).
    end_time : float
        The time the time series end (in seconds) relative to the original time
        reference.
    unit : str
        String containing the name of the unit for the data (for plotting).
    nsamples : int
        Number of samples per time series in the data array.
    nfeatures
        Number of features for each sample in the data array.
    unit
        Unit for the data (for plotting).
    caption
        Text caption.
    label
        Label for the time series.
    """

    # these properties should be overriden by the subclass if they are supposed
    # to be different
    unit = ''
    caption = ''
    label = ''

    def __init__(self, fs, data=None, start_time=0., unit=None):
        """
        """
        if fs <= 0:
            raise ValueError(
                "the sampling frequency (fs) must be greater than zero")

        self.data = None

        self.fs = np.float_(fs)
        self.start_time = 0 if start_time is None else np.float_(start_time)

        if unit:
            self.unit = unit

        if data is not None:
            self._write_data(data)

    @property
    def nsamples(self):  # pylint: disable=missing-docstring
        return self.data.shape[-1]

    @property
    def nfeatures(self):  # pylint: disable=missing-docstring
        if self.data.ndim == 1:
            return 1
        elif self.data.ndim == 2:
            return self.data.shape[-2]

    @property
    def duration(self):  # pylint: disable=missing-docstring
        return self.nsamples / self.fs

    @property
    def nyquist(self):  # pylint: disable=missing-docstring
        return self.fs / 2

    @property
    def end_time(self):  # pylint: disable=missing-docstring
        return self.start_time + self.duration

    @property
    def ts(self):  # pylint: disable=missing-docstring
        return 1 / self.fs

    @property
    def time(self):  # pylint: disable=missing-docstring
        start = self.start_time
        end = self.end_time
        return np.linspace(start, end, self.nsamples)

    def copy(self):
        "Return a copy of the time series object (deep copy)."
        return cp.deepcopy(self)

    def normalize(self):
        "Return a normalized copy of the time series."
        normalized_data = self.data / np.max(self.data)

        ts = cp.copy(self)
        ts.data = normalized_data
        return ts

    def diff(self, n=1):
        "Return the n-th discrete difference for the time series"
        nfeatures = self.nfeatures
        dtype = self.data.dtype
        data = np.reshape(self.data, (nfeatures, -1))

        zero_pre_pad = np.zeros((nfeatures, n), dtype)
        padded_data = np.concatenate((zero_pre_pad, data), axis=-1)
        data_diff = np.diff(padded_data, n, axis=-1)

        if nfeatures == 1:
            data_diff.shape = (self.nsamples, )

        ts = cp.copy(self)
        ts.data = data_diff

        return ts

    def zeros_to_nan(self):
        """
        Converts zeros to np.nan in the data array. Returns a new time series.
        """
        ts = self.copy()
        ts.data[self.data == 0] = np.nan
        return ts

    def hwr(self):
        "Return a half-wave rectified copy of the time series."
        rectified_data = np.clip(self.data, 0, None)

        ts = cp.copy(self)
        ts.data = rectified_data

        return ts

    def plot(self):
        "Plot the time series using matplotlib."
        f = plt.figure(figsize=(15, 9))
        plt.plot(self.time, self.data, label=self.label)
        if self.label:
            plt.legend(loc='lower right', ncol=2, fontsize='x-small')
        plt.title(self.caption)
        plt.ylabel(self.unit)
        plt.xlabel('time (s)')
        plt.show()

        return f

    def time_to_sample_index(self, time):
        """
        Convert time (in seconds) to the correspoding sample index in the time
        series.
        """
        time = time + self.start_time
        index = round(time * self.fs)
        return index

    def __repr__(self):
        """Representation for TimeSeries object."""
        class_name = self.__class__.__name__
        dim = '(nfeatures={}, nsamples={})'.format(self.nfeatures,
                                                   self.nsamples)
        other = 'fs={}, unit={}, label={}'.format(self.fs, self.unit,
                                                  self.label)
        return '{}: {}, {}'.format(class_name, dim, other)

    def __getitem__(self, sl):
        """
        Get an excerpt from the time series using slices. Return a new
        TimeSeries object.
        """
        if type(sl) == Segment:
            time_offset = conversion.sample_index_to_seconds(
                sl.start, sl.fs)
            sl = sl.generate_slice(self)
        elif (type(sl) == slice):
            index_start = sl.start or sl.stop
            time_offset = conversion.sample_index_to_seconds(
                index_start, self.fs)
        else:
            raise ValueError("invalid value for slicing operation: must be " +
                             "of type `Segment` or a Python slice")

        sliced_data = self.data[sl]
        ts = cp.copy(self)
        ts.data = sliced_data
        ts.start_time += time_offset  # shift start
        return ts

    def get_samples(self, start, stop):
        """
        Get an excerpt from the time series using sample indexes. Return a new
        TimeSeries object.
        """
        return self[..., start:stop]

    # Arithmetic, relational and boolean operations
    def __add__(self, other):
        """Add two time series."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data + other.data
        return ts

    def __sub__(self, other):
        """Subtract two time series."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data - other.data
        return ts

    def __mul__(self, other):
        """Multiplicate two time series element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data * other.data
        return ts

    def __truediv__(self, other):
        """Divide two time series element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data / other.data
        return ts

    def __mod__(self, other):
        """Division remainder for two time series taken element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data % other.data
        return ts

    def __lt__(self, other):
        """Less than"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data < other.data
        return ts

    def __le__(self, other):
        """Less than or equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data <= other.data
        return ts

    def __gt__(self, other):
        """Greater than"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data > other.data
        return ts

    def __ge__(self, other):
        """Greater than or equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data >= other.data
        return ts

    def __eq__(self, other):
        """Equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data == other.data
        return ts

    def __ne__(self, other):
        """Not equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data != other.data
        return ts

    def __len__(self):
        """Length of the time series -- number of samples."""
        return self.data.shape[-1]

    def _write_data(self, data):
        """
        Write the given data to the data field. This method should be
        implemented by subclasses to control how the data field is initialized.

        Be careful using this method; it modifies the object.
        """
        self.data = data

    def _shift_start(self, seconds):
        """
        Shift the start of the time series by the number of seconds given.

        Be careful using this method; it modifies the object.
        """

        self.start_time += seconds


class Audio(TimeSeries):
    """
    Stores audio data, which can be loaded from a file or an array.

    Parameters
    ----------
    filename : str, optional
        Name of the audio file to be loaded.
    data : np.array
        Data vector containing the audio data to be loaded.
    fs : int, optional
        Sampling frequency of the data.
    caption : str, optional
        caption for the audio file loaded (optional).

    Examples
    --------
    There are two different ways to initialize an Audio object: from
    audio files or from NumPy arrays.

    To initialize it using an audio file, you just need to pass the
    ``filename`` to be loaded:

    >>> a1 = Audio('clarinet.wav')

    To initialize it using a NumPy array, two arguments are necessary: the
    ``data`` array and the sampling frequency ``fs``:

    >>> a2 = Audio(clarinet_data, 44100)

    There is also the optional parameter ``caption`` which is a
    textual description used for plotting and displaying reports
    about the audio file. In case you load the audio from a file
    and don't specify a caption, the filename will be assigned to
    it.
    """
    def __init__(self, *args, **kwargs):
        """
        Constructor for Audio class.
        """
        self.unit = 'amplitude'
        self.label = 'waveform'

        nargs = len(args)
        caption = kwargs.get('caption', None)

        # one argument: file name
        if nargs == 1:
            filename = args[0]
            filename = path.expanduser(filename)  # expanding ~ to /home/user

            # loading audio file
            data, fs = read(filename)
            self.filename = path.basename(filename)
            self.caption = caption or self.filename

        # two arguments: an array and a sampling frequency
        elif nargs == 2:
            data, fs = args[0], args[1]
            self.filename, self.caption = None, caption

        else:
            raise (TypeError(
                'invalid number of positional arguments: should be '
                '1 or 2'))

        super(Audio, self).__init__(fs, data=data, unit=self.unit)

    def play(self):
        """
        Play audio from Audio object.
        """
        player.play(self)

    def play_from_time(self, from_time):
        """
        Play audio from Audio object start at time ``from_time``.
        """
        player.play_interval_seconds(self, from_time, None)

    def play_segment(self, segment):
        """
        Play segment from Audio obejct.
        """
        player.play_interval_seconds(self, segment.start_time, segment.end_time)

    def stop(self):
        """
        Stop playing audio.
        """
        player.stop()


class DimensionalityError(Exception):
    """
    Exception raised for errors in dimensionality of arays
    """
    pass



"""
GUI for audio annotation.
"""

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Button, MultiCursor # pylint: disable=import-error

import iracema
import iracema.segment as segment
import iracema.features


class NoteAnnotator:
    """
    GUI for annotating note onsets and offsets for audio files.
    """

    def __init__(self, audio_file, output_csv, input_csv=None,
                 start_only=False):
        # loading audio file
        self.audio_file = audio_file
        self.audio = iracema.Audio(audio_file).normalize()

        # initializing segment list
        if input_csv:
            self.notes = segment.SegmentList.load_from_csv_file(
                self.audio, input_csv, limits='seconds', start_only=start_only)
        else:
            self.notes = segment.SegmentList([])

        self.output_csv = output_csv

        # playhead, onset_cursor and segment_spans
        self.fps = 10.
        self.i_playhead = 0
        self.playhead_positions = np.linspace(self.audio.start_time,
                                              self.audio.end_time,
                                              self.audio.duration * self.fps)
        self.onset_cursor = None
        self.segment_spans = []

        # state (boolean)
        self.state = 'waiting'  # {'waiting', 'playing', 'editing', 'editing-playing'}

        # plot waveform axes
        self.f = plt.figure(figsize=(16, 9), dpi=100)
        self.ax_waveform = self.f.add_subplot(211, picker=1)
        self.f.subplots_adjust(left=0.05, right=1 - 0.05)

        self.plot_waveform(self.audio)
        self.ax_waveform.set_xlim(self.audio.start_time, self.audio.end_time)
        self.ax_waveform.set_gid('ax1')

        # plot playhead and segment spans
        self.playhead = self.ax_waveform.axvline(
            x=self.playhead_positions[self.i_playhead], color='black')

        self.update_segment_spans()

        # create buttons
        ax_but_add_onset = self.f.add_subplot(425)
        self.but_add_onset = Button(ax_but_add_onset, 'add note onset')

        ax_but_add_offset = self.f.add_subplot(426)
        self.but_add_offset = Button(ax_but_add_offset, 'add note offset')

        ax_but_save = self.f.add_subplot(817)
        self.but_save = Button(ax_but_save, 'save to csv')

        ax_but_play = self.f.add_subplot(818)
        self.but_play = Button(ax_but_play, 'play / stop\n[space]')

        # connecting events
        self.but_play.on_clicked(self.play_or_stop)
        self.but_save.on_clicked(self.save_csv)
        self.but_add_onset.on_clicked(self.add_onset)
        self.but_add_offset.on_clicked(self.add_offset)
        self.f.canvas.mpl_connect('key_press_event',
                                  self.on_keypress)  # key presses
        self.f.canvas.mpl_connect('pick_event', self.on_pick)  # pick events

        # animation
        self.anim = animation.FuncAnimation(
            self.f, self.forward_playhead, interval=1000 / self.fps)

        # display figure
        self.f.canvas.manager.full_screen_toggle()

        plt.show()

    # axes plots
    def plot_waveform(self, audio):
        "Plot the waveform."
        rms = iracema.features.rms(audio, 2048, 1024)  # TODO: fix hard code
        self.ax_waveform.plot(
            rms.time, rms.data, linewidth=1.2, alpha=0.9, color='blue')

    def plot_spectrogram(self, ax, spectrogram):
        "Plot the spectrogram."
        pass

    # callbacks
    def on_pick(self, event):
        print("artist: {} - {}, coords: {},{}".format(event.artist,
                                                      event.artist.get_gid(),
                                                      event.mouseevent.xdata,
                                                      event.mouseevent.ydata))

        if event.artist.get_gid() == 'ax1':
            if self.state == 'playing':
                self.audio.stop()
                self.state = 'waiting'
            self.update_playhead(event.mouseevent.xdata)

    def on_keypress(self, event):
        if event.key == ' ':
            self.play_or_stop(event)
        elif event.key == 'enter':
            self.stop_playing()
            self.reset_playhead()

    def add_onset(self, event):
        if self.state == 'waiting':
            self.state = 'editing'
            print('add onset on: ', self.playhead.get_xdata())
            self.onset_cursor = self.ax_waveform.axvline(
                x=self.playhead.get_xdata(), color='red', linewidth=2)

    def add_offset(self, event):
        if self.state == 'editing':
            self.state = 'waiting'
            onset = self.onset_cursor.get_xdata()[0]
            offset = self.playhead.get_xdata()
            self.onset_cursor.remove()

            new_note = segment.Segment(
                self.audio, onset, offset, limits='seconds')
            self.notes.append(new_note)
            self.update_segment_spans()

    def update_segment_spans(self):
        for span in self.segment_spans:
            span.remove()

        self.segment_spans = []

        for note in self.notes:
            note_span = self.ax_waveform.axvspan(
                xmin=note.start_time,
                xmax=note.end_time,
                edgecolor='black',
                facecolor='green',
                alpha=0.25,
                linewidth=1)
            self.segment_spans.append(note_span)

    def play_or_stop(self, event):
        if self.state == 'waiting':
            self.state = 'playing'
            self.audio.play_from_time(self.playhead_positions[self.i_playhead])
            plt.show()
        elif self.state == 'playing':
            self.state = 'waiting'
            self.audio.stop()
        elif self.state == 'editing':
            self.state = 'editing-playing'
            self.audio.play_from_time(self.playhead_positions[self.i_playhead])
        elif self.state == 'editing-playing':
            self.state = 'editing'
            self.audio.stop()

    def stop_playing(self):
        if self.state == 'playing':
            self.state = 'waiting'
            self.audio.stop()

    def reset_playhead(self):
        self.update_playhead(0)

    def update_playhead(self, x):
        self.playhead.set_xdata(x)
        self.i_playhead = int(x * self.fps)

    def forward_playhead(self, _):
        if self.state == 'playing' or self.state == 'editing-playing':
            self.i_playhead += 1
            # check if the playhead has reached the end of the audio
            if self.i_playhead == self.playhead_positions.shape[-1]: # pylint: disable=no-member
                self.i_playhead = 0
                self.state = 'waiting'
            self.playhead.set_xdata(self.playhead_positions[self.i_playhead])

    def save_csv(self, _):
        self.notes.save_to_csv_file(self.output_csv)



"""
Methods for loading and writing data to CSV files.
"""


# TODO: this code is outdated, should be rewritten
def load_from_csv_file(time_series,
                       filename,
                       limits='seconds'):
    """
    Load info from a csv file into a list.
    """
    import csv
    start, end = [], []

    with open(filename, 'r', newline='') as ff:
        for row in csv.reader(ff, delimiter=','):
            start.append(np.float_(row[0]))
            if not start_only:
                end.append(np.float_(row[1]))

    segments = [
        cls(time_series, st, end, limits) for st, end in zip(start, end)
    ]
    return SegmentList(segments)


def save_to_csv_file(self, filename, limits='seconds'):
    """
    Save segment list to a csv file.
    """
    import csv

    with open(filename, 'w', newline='') as f:
        writer = csv.writer(f, delimiter=',')
        for note in self:
            if limits == 'seconds':
                line = [note.start_time, note.end_time]
            else:
                line = [note.start, note.end]

            if start_only:
                writer.writerow([line[0]])
            else:
                writer.writerow(line)




from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='iracema',
    version='0.1.3',
    url='http://github.com/cegeme/iracema',
    author='Tairone Magalhães',
    author_email='taironemagalhaes@gmail.com',
    description='Audio Content Analysis for Research on Musical Expressiveness and Individuality',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    install_requires=[
        'numpy>=1.18.4',
        'scipy>=1.4.1',
        'sounddevice>=0.3.12',
        'audioread>=2.1.8',
        'matplotlib==3.2.1'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(),
    )
